/********************************************************************
 * cl_utils.c
 * Part of maKs
 * Copyright Chris Benndorf, 1998
 * Released under GNU Public Library License
 * Best readeable with ":set tabstop=4"
 *
 * $Id: cl_utils.c,v 1.5 2000/01/13 21:50:32 root Exp root $
 *
 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <glob.h>
#include <log.h>
#include <cl_utils.h>

cl_mgmt *clptr;

/********************************************************************
 * Allocate space for client struct and initialize
 * returns pointer to new client struct or NULL 
 ********************************************************************/
cl_mgmt *Cl_create()
{
	cl_mgmt *ret;
	
	ret=(cl_mgmt *)calloc(sizeof(cl_mgmt),1);
	if (ret == NULL) {
		logerror("create client");
		exit(1);
	}

	ret->fd		= -1;
	ret->magic	= 0xAFFEAFFE;
	ret->next	= NULL;
	ret->prev	= NULL;

	return(ret);
}

/********************************************************************
 * Add given client to list
 * returns pointer to added client struct
 ********************************************************************/
cl_mgmt *Cl_add(cl_mgmt *this)
{
	cl_mgmt *prev;

	prev = Cl_last();

	prev->next = this;
	this->prev = prev;
	this->next = NULL;

	return(this);
}

/********************************************************************
 * Set fields in client struct
 * returns pointer to client struct
 ********************************************************************/
cl_mgmt *Cl_init(cl_mgmt *this, int fd, WORD ctyp, char *appl, char *user)
{
	this->fd		= fd;
	this->magic		= 0xAFFEAFFE;
	this->idleProc 	= NULL;
	this->readProc 	= NULL;
	this->writeProc	= NULL;
	this->closed 	= 0;
	this->accepted 	= 0;
	this->clinkid 	= 0;
	this->rndid 	= 0;
	this->flgnr 	= 0;
	this->inval 	= 0;
	this->pcknr 	= 0x7E;
	this->perm 	= 0;
	this->ctyp 	= ctyp;
        this->read_buf.size	= 0;
        this->read_buf.pos	= 0;
        this->read_buf.ptr	= NULL;
        
                
	strncpy(this->appl,appl,MAX_APPSTR_LEN);
	strncpy(this->user,user,MAX_USRSTR_LEN);

	return(this);
}

/********************************************************************
 * Delete given client from list
 * returns pointer to previous client struct
 * (There must be a prev since server and serial don't get deleted
 ********************************************************************/
cl_mgmt *Cl_delete(cl_mgmt *this)
{
	cl_mgmt *prev;
	cl_mgmt *next;


	prev = this->prev;
	next = this->next;

	if (next != NULL) {
		prev->next = this->next;
		next->prev = this->prev;
	} else 
		prev->next = NULL;

        if (this->read_buf.ptr) {
                free (this->read_buf.ptr);
        }
	free(this);
	
	return(prev);
}

/********************************************************************
 * Get serial port client (always second client)
 * returns pointer to client struct
 ********************************************************************/
cl_mgmt *Cl_find_serial()
{
	cl_mgmt *this;

	this = clptr;
	this = this->next;
	return(this);
}

/********************************************************************
 * Get client with given linkid
 * returns pointer to client struct or NULL if not found
 ********************************************************************/
cl_mgmt *Cl_find_by_lid(WORD lid)
{
	cl_mgmt *this;

	this = clptr;

	while (this != NULL) {
		if ((lid) && (this->clinkid == lid))
			return(this);
		this = this->next;
	}

	return(NULL);
}

/********************************************************************
 * Get client with given linkid
 * returns pointer to client struct or NULL if not found
 ********************************************************************/
cl_mgmt *Cl_find_by_rndid(BYTE rndid)
{
	cl_mgmt *this;

	this = clptr;

	while (this != NULL) {
		if ((this->rndid == rndid) && (this->rndid))
			return(this);
		this = this->next;
	}

	return(NULL);
}

/********************************************************************
 * Find next client without valid linkid
 * returns pointer to client struct or NULL if not found
 ********************************************************************/
cl_mgmt *Cl_find_invalid()
{
	cl_mgmt *this;

	this = clptr;

	while (this != NULL) {
		if (Cl_is_app(this) && this->inval)
			return(this);
		this = this->next;
	}

	return(NULL);
}

/********************************************************************
 * Get n'th client struct in list
 * returns pointer to n'th client struct or NULL if idx out of range
 ********************************************************************/
cl_mgmt *Cl_index(int idx)
{
	cl_mgmt *this;
	int		i = 1;

	this = clptr;

	while (i < idx) {
		if (this->next != NULL) {
			this = this->next;
		} else { 
			return(NULL);
		}
		i++;
	}

	return(this);
}

/********************************************************************
 * Get last client struct in list
 * returns pointer to last client struct
 ********************************************************************/
cl_mgmt *Cl_last()
{
	cl_mgmt *this;

	this = clptr;

	while (this->next != NULL) {
		this = this->next;
	}

	return(this);
}

/********************************************************************
 * Get first client struct in list
 * returns pointer to last client struct
 ********************************************************************/
cl_mgmt *Cl_first()
{
	return(clptr);
}

/********************************************************************
 * Checks if this client has sent a logout
 * returns boolean
 ********************************************************************/
int Cl_is_closed(cl_mgmt *this)
{
	return(this->closed != 0);
}

/********************************************************************
 * Checks if this client is a server control application
 * returns boolean
 ********************************************************************/
int Cl_is_ctrl(cl_mgmt *this)
{
	return(this->ctyp == CTI_SVRCTRL);
}

/********************************************************************
 * Checks if this client is an application
 * returns boolean
 ********************************************************************/
int Cl_is_app(cl_mgmt *this)
{
	return(this->ctyp >= CTI_CL_APP);
}

/********************************************************************
 * Checks if this client is an application
 * returns boolean
 ********************************************************************/
int Cl_needs_BCast(cl_mgmt *this)
{
	int needsbc = ((this->perm & PERM_BMON)  == PERM_BMON) ||
				  ((this->perm & PERM_S0MON) == PERM_S0MON) ||
				  ((this->perm & PERM_CTI)   == PERM_CTI);
	return(needsbc);
}

/********************************************************************
 * FDSETs all fds in list
 * returns last client in list
 ********************************************************************/
cl_mgmt *Cl_fdset_all(fd_set *set)
{
	cl_mgmt *this;

	this=clptr;

	while (this != NULL) {
		if (this->fd != -1)
			FD_SET(this->fd,set);

		this = this->next;
	}
	return(this);
}

/********************************************************************
 * finds highest fd in client list
 * returns highest fd in list
 ********************************************************************/
int Cl_highfd()
{
	cl_mgmt *this;
	int	highest	= 0;

	this = clptr;

	while (this != NULL) {
		if ((this->fd > highest))
			highest = this->fd;

		this = this->next;
	}

	return(highest + 1);
}

/********************************************************************
 * sets last_pak of all connected clients
 * returns last client struct
 ********************************************************************/
cl_mgmt *Cl_touch_all()
{
	cl_mgmt *this;
	int	i;

	this = clptr;

	while (this != NULL) {
		if ((this->fd != -1) && Cl_is_app(this)) {
			this->last_pak = time(NULL);
		}

		this = this->next;
	}
	return(this);
}

/********************************************************************
 * sends buffer to all connected SRVR_CTRLs
 * returns last client struct
 ********************************************************************/
cl_mgmt *Cl_send_all_ctrl(void *buf, int len)
{
	cl_mgmt *this;
	int	i;

	this = clptr;

	while (this != NULL) {
		if ((this->fd != -1) && (this->writeProc) && Cl_is_ctrl(this)) {
			this->err = (this->writeProc)(this, buf, len);
			this->last_pak = time(NULL);
		}

		this = this->next;
	}
	return(this);
}

/********************************************************************
 * sends buffer to all connected clients
 * returns last client struct
 ********************************************************************/
cl_mgmt *Cl_send_all(void *buf, int len)
{
	cl_mgmt *this;
	int	i;

	this = clptr;

	while (this != NULL) {
		if ((this->fd != -1) && (this->writeProc) && Cl_is_app(this) && Cl_needs_BCast(this)) {
			this->err = (this->writeProc)(this, buf, len);
			this->last_pak = time(NULL);
		}

		this = this->next;
	}
	return(this);
}

/********************************************************************
 * sends buffer to specified client
 * returns last client struct
 ********************************************************************/
cl_mgmt *Cl_send(cl_mgmt *this, BYTE *buf, int len)
{
	if (this != NULL) {
		if ((this->fd != -1) && (Cl_is_app(this))) {
			(this->writeProc)(this, buf, len);
			this->last_pak = time(NULL);
		}

	}
	return(this);
}

/********************************************************************
 * Checks if user is allowed to use this funtionality
 * returns Bool (int)
 ********************************************************************/
int Cl_checkUser(cl_mgmt *this)
{
	/* Not implemented yet */
	return(1);
}

/********************************************************************
 * sends buffer to all connected clients
 * returns last client struct
 ********************************************************************/
cl_mgmt *Cl_accept(int fd)
{
	cl_mgmt	*cl;
	int	msgsock;
	int	s_addr;

	cl = Cl_create();

	s_addr = sizeof(struct sockaddr);

	msgsock = accept(fd,&cl->addr, &s_addr);
	if (msgsock != -1) {
		Cl_init(cl,msgsock,CTI_CL_TELNET,"unknown","unknown");
		Cl_add(cl);
		return(cl);
	} else 
		return(NULL);
}

/********************************************************************
 * Prints out status of a client
 * returns the client
 ********************************************************************/
cl_mgmt *Cl_print(cl_mgmt *this, FILE *fp)
{
	LONG    *ap = (LONG *)&this->addr.sa_data[2];

	if ((this != NULL)) {
		logprintf(LOG_MSG,
			"Socket : %02X\t"
			"Open   : %s\n"
			"LinkID : %04X\t"
			"Random : %02X\n"
			"Typ    : %02X\t"
			"Perm   : %02X\n"
			"PackNr : %02X\t"
			"FolgNr : %02X\n"
			"ResvTl : %02X\t"
			"Invalid: %02X\n"
			"P. sent: %05d\t"
			"P. rcvd: %05d\n"
			"E. sent: %05d\t"
			"E. rcvd: %05d\n"
			"User   : %s\n"
			"Appl   : %s\n"
			"IP-Addr: %s\n"
			,
			this->fd,
			(this->closed) ? "No" : "Yes",
			this->clinkid,
			this->rndid,
			this->ctyp,
			this->perm,
			this->pcknr,
			this->flgnr,
			this->tln_res,
			this->inval,
			this->pkg_sent,
    		this->pkg_rcvd,
    		this->pkg_serr,
    		this->pkg_rerr,
			this->user,
			this->appl,
			inet_ntoa(*ap)
		);
	}

	return(this);
}

/********************************************************************
 * Prints out status of all clients
 * returns last client in list
 ********************************************************************/
cl_mgmt *Cl_status(FILE *fp)
{
	cl_mgmt *this;
	int i = 0;

	this = clptr;

	logdate(LOG_MSG);
	while ((this != NULL)) {

		logprintf(LOG_MSG, "\nClient#: %d\n" , i);

		Cl_print(this, fp);

		if (this->magic != 0xAFFEAFFE) break; 
		this = this->next;
		i++;
	}
	return(this);
}

/********************************************************************
 * Builds string with status of client
 * returns string
 ********************************************************************/
char *Cl_sstatHead(int typ)
{
	switch (typ) {
		case 0x00 : return((char *)strdup(ADDR_INFO_HEADER));
		case 0x01 : return((char *)strdup(NORM_INFO_HEADER));
		case 0x02 : return((char *)strdup(MGMT_INFO_HEADER));
		case 0x04 : return((char *)strdup(TECH_INFO_HEADER));
		case 0x80 : return((char *)strdup("BUTTONINFO"));
		default:	return((char *)strdup(UNDEF_INFO_HEADER));
	}
}

/********************************************************************
 * Builds string with status of client
 * returns string
 ********************************************************************/
char *Cl_strstatus(int idx, int typ)
{

#define TIME_MAX_LEN 30
	cl_mgmt *this;
	char	ret[128];
	char	name[20];
	char	p_time[TIME_MAX_LEN];
	char	*p;
	char *conTypes[]= {
   					"NOINIT",
    				"LISTEN",
    				"SERIAL",
    				"SRVR_CNTRL",
    				"PROGRAM",
    				"TELNET"
	}; 
	char *buttonMsg[]= {
   					"IP/Name",
    				"Rechte",
    				"LCB",
    				"Pakete"
	}; 
	const size_t bmsgsize = sizeof(buttonMsg)/sizeof(char*);

	if (idx == 0 && typ !=0x80) return(Cl_sstatHead(typ));

	if ((this=Cl_index(idx))==NULL) return(NULL);

	/* Get rid of some redundant Euraconf info */
	if (this->appl[0]=='[')	{
		p = (char *)strrchr(this->appl,']');
		p++;
	} else {
		p = this->appl;
	}
	
	if (!strncmp(p,"Euraconf",8)) {
		p = (char *)strchr(p,' ');
		p++;
	}
	/* End of special Euraconf handling */

	if (strlen(p) > 14) {
		strncpy(name, p, 14);
		name[14]=0;
	} else {
		strcpy(name, p);
	}
	switch (typ) {
		case 0x00 :
			sprintf(ret, ADDR_INFO_FORMAT, 
				this->clinkid, this->user, this->appl,
				(BYTE)this->addr.sa_data[2],
				(BYTE)this->addr.sa_data[3],
				(BYTE)this->addr.sa_data[4],
				(BYTE)this->addr.sa_data[5]
			);
			break;
		case 0x01 :
			strftime(p_time, TIME_MAX_LEN, "%d.%m.%y %H:%M", localtime(&this->login));
			sprintf(ret, NORM_INFO_FORMAT, 
				this->clinkid, this->user, name, conTypes[this->ctyp],
				this->perm, this->tln_res, p_time);
			break;
		case 0x02 :
			sprintf(ret, MGMT_INFO_FORMAT, 
				this->clinkid, this->user, name,
				this->rndid, this->pcknr, this->flgnr, this->fd);
			break;
		case 0x04 :
			sprintf(ret, TECH_INFO_FORMAT, 
				this->clinkid, this->user, name,
				this->pkg_sent, this->pkg_rcvd, this->pkg_serr, this->pkg_rerr, this->pkg_left, this->fd);
			break;
		case 0x80 :
			if (idx < bmsgsize) {
				sprintf(ret,"%s",buttonMsg[idx]);
			} else {
				ret[0]='\0';
			}
			break;
		default:
			ret[0]='\0';
	}

	return((char *)strdup(ret));
}

/********************************************************************
 * Close and delete client
 ********************************************************************/
void Cl_close(cl_mgmt *this)
{
	close(this->fd);
	this->closed	= 1;
	this->fd		= -1;
}

/********************************************************************
 * Find out which clients need to be handled
 ********************************************************************/
void Cl_new_client(cl_mgmt *this)
{
	cl_mgmt *cl;

	cl = Cl_accept(this->fd);
	
	if (cl)
		clnt_init(cl);

	this->pkg_rcvd++;
}

/********************************************************************
 * Find out which clients need to be handled
 ********************************************************************/
void Cl_handle_data(fd_set *fds)
{
	cl_mgmt *this;

	this = clptr;

	while (this != NULL) {

		if ((this->fd == -1) || (FD_ISSET(this->fd, fds))) {
			switch (this->ctyp){
				case CTI_CL_LISTEN :
					Cl_new_client(this);
					break;
				case CTI_CL_SERIAL :
					l3_handle_pbx_data(this);
					break;
				case CTI_CL_TELNET :
				case CTI_SVRCTRL :
				case CTI_CL_APP :
					if (this->fd != -1)
						l3_handle_cl_data(this);
					if (Cl_is_closed(this)) {
						this = Cl_delete(this);
					}
					break;
				default:
					break;
			}
		}

		this = this->next;
	}
}

/********************************************************************
 * Call idleProc for every client
 ********************************************************************/
void Cl_idle()
{
	cl_mgmt *this;

	/* this = clptr;*/
	this = Cl_last();

	while (this != NULL) {

/*
		if (this->inval)  {
			l3_linkid_req(this);
			return;
		}
*/
			
		if (this->idleProc)
			(this->idleProc)(this);

		/* this = this->next; */
		this = this->prev;
	}
}
