/********************************************************************
 * level2clnt.c
 * Part of maKs
 * Copyright Chris Benndorf, 1998
 * Released under GNU Public Library License
 * Best readeable with ":set tabstop=4"
 *
 * $Id: level2clnt.c,v 1.4 1999/12/30 21:00:15 root Exp root $
 *
 ********************************************************************/
 
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <glob.h>
#include <log.h>
#include <cl_utils.h>
#include <level3.h>

#ifndef TEMP_FAILURE_RETRY
#define TEMP_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int __result;                                                     \
       do __result = (long int) (expression);                                 \
       while (__result == -1L && errno == EINTR);                             \
       __result; }))
#endif

extern FILE *logfp;

void clnt_idle(cl_mgmt *this) 
{
	time_t now;
	
	now = time(NULL);

	if (Cl_is_app(this) && (!this->clinkid) && (this->flgnr == 0x02) && (this->last_pak + 30 < now)) {
			this->last_pak = now; /* wait another 30 secs */
			logprintf(LOG_MSG, "LINK_ID(GET) for Client #%d seems to have failed... Requesting new one!\n",this->fd);
			l3_linkid_req(this);
			return;
	}

	if (Cl_is_app(this) && (this->last_pak + 30 < now)) {
		logprintf(LOG_MSG, "Client #%d is 30 secs without packet...\n",this->fd);
		this->last_pak = now; /* wait another 30 secs */

		if (Cl_is_closed(this))
			return;

		if ((this->clinkid))
			l3_verfylink_get(this);
	}
}

int clnt_write(cl_mgmt *cl, BYTE *buf, int len)
{
	int ret;

	logbuf(LOG_CLNTBYTES, "clnt_write", buf, len);
        /* is this blocking? if yes, we have to find a non-blocking
         * way!
         */
	ret = TEMP_FAILURE_RETRY (write (cl->fd, buf, len));

	if (ret != len) {
		ret = -1;
		cl->pkg_serr++;
	} else {
		ret = 0;
		cl->pkg_sent++;
	}

	return(ret);
}

/* clnt_read - read (part of) packet from client, data bytes are
 *             stored inside of cl until packet completion.
 *
 * IN:      cl	= Client willing to send a packet
 * OUT:     len	= 0, if packet incomplete by now
 *                else length of complete packet
 * RETURNS: NULL on error, else pointer to buffer
 */
BYTE *clnt_read(cl_mgmt *cl, int *len)
{
	ssize_t bytes_read;
	size_t packet_size = 0;
	BYTE *complete_packet = NULL;
	BYTE *new_ptr = NULL;
	struct io_buffer *buf = &cl->read_buf;

	*len = 0; /* mark the packet incomplete */
	if (buf->ptr == NULL) {
		/* new packet to assemble */
		buf->size = 5;
		if ((buf->ptr = malloc (buf->size)) == NULL) {
			return (NULL);
		}
		buf->pos = 0;
	}

	if (buf->pos < 5) {                
		if ((bytes_read = TEMP_FAILURE_RETRY (read (cl->fd, buf->ptr, buf->size))) <= 0) {
			/* NOTE: reading ZERO bytes is very uncommon
			* here, since select() told us there is data!
			*/
			if (bytes_read == 0) {
				logprintf (LOG_CLNTBYTES, "clnt_read: receive error\n");
				cl->pkg_rerr++;
			}
			return (NULL);
		}
		buf->pos += bytes_read;
	}

	logprintf (LOG_CLNTBYTES, "read1 %d(%d) bytes...\n", bytes_read, buf->pos);
	if (buf->pos >= 5) {
		packet_size = 5 + (size_t)(buf->ptr[4] & 0xff);
		if (buf->size < packet_size) {
			/* we have to allocate buffer space for the
			* rest of the packet
			*/
			if ((new_ptr = realloc (buf->ptr, packet_size)) == NULL) {
				return (NULL);
			}
			buf->ptr = new_ptr;
		}
		if (buf->pos < packet_size) {
			/* packet incomplete... */
			if ((bytes_read = TEMP_FAILURE_RETRY (read (cl->fd, &buf->ptr[buf->pos], packet_size - 5))) < 0) {
				return (NULL);
			}
			logprintf (LOG_CLNTBYTES, "read2 %d(%d) bytes...\n", bytes_read, buf->pos);
			buf->pos += bytes_read;
		}
		if (buf->pos == packet_size) {
			/* packet complete */
			*len = packet_size;
			complete_packet = buf->ptr;
			buf->ptr = NULL;
			cl->pkg_rcvd++;
			logbuf(LOG_CLNTBYTES, "clnt_read", complete_packet, packet_size);
			return (complete_packet);
		}
	}

	logprintf (LOG_CLNTBYTES, "packet incomplete... delayed (read %d/%d bytes)\n", buf->pos, buf->size);
	return (buf->ptr);

}

void clnt_init(cl_mgmt *cl)
{
    cl->idleProc    = &clnt_idle;
    cl->readProc    = &clnt_read;
    cl->writeProc   = &clnt_write;
}
