/* ******************************************************************
 * Q931.cpp
 * Part of maKs
 * Copyright Chris Benndorf, 1998
 * Released under GNU Public Library License
 * Best readeable with ":set tabstop=4"
 *
 * $Id$
 *
 * Most of this is stolen from Karsten Keils HiSax-Driver...
 * ******************************************************************/

#include <iostream>
#include <strstream>
#include <iomanip>
#include <Q931.h>

static struct MessageType {
  		u_char nr;
  		char *descr ;
} mtlist[] = {
  		{ 0x0, "UNKNOWN" },
  		{ 0x1, "ALERTING" },
  		{ 0x2, "CALL PROCEEDING" },
  		{ 0x7, "CONNECT" },
  		{ 0xf, "CONNECT ACKNOWLEDGE" },
  		{ 0x3, "PROGRESS" },
  		{ 0x5, "SETUP" },
  		{ 0xd, "SETUP ACKNOWLEDGE" },
  		{ 0x26, "RESUME" },
  		{ 0x2e, "RESUME ACKNOWLEDGE" },
  		{ 0x22, "RESUME REJECT" },
  		{ 0x25, "SUSPEND" },
  		{ 0x2d, "SUSPEND ACKNOWLEDGE" },
  		{ 0x21, "SUSPEND REJECT" },
  		{ 0x20, "USER INFORMATION" },
  		{ 0x45, "DISCONNECT" },
  		{ 0x4d, "RELEASE" },
  		{ 0x5a, "RELEASE COMPLETE" },
  		{ 0x46, "RESTART" },
  		{ 0x4e, "RESTART ACKNOWLEDGE" },
  		{ 0x60, "SEGMENT" },
  		{ 0x79, "CONGESTION CONTROL" },
  		{ 0x7b, "INFORMATION" },
  		{ 0x62, "FACILITY" },
  		{ 0x6e, "NOTIFY" },
  		{ 0x7d, "STATUS" },
  		{ 0x75, "STATUS ENQUIRY" }
} ;


static struct InformationElement {
  		u_char nr;
  		char *descr;
} const ielist[] = {
  		{ 0x00, "Segmented message"},
  		{ 0x04, "Bearer capability"},
  		{ 0x08, "Cause"},
  		{ 0x10, "Call identity"},
  		{ 0x14, "Call state"}, 
  		{ 0x18, "Channel identification"},
  		{ 0x1c, "Facility"}, 
  		{ 0x1e, "Progress indicator"},
  		{ 0x20, "Network-specific facilities"},
  		{ 0x27, "Notification indicator"},
  		{ 0x28, "Display"},
  		{ 0x29, "Date/Time"},
  		{ 0x2c, "Keypad facility"},
  		{ 0x34, "Signal"},
  		{ 0x40, "Information rate"},
  		{ 0x42, "End-to-end delay"},
  		{ 0x43, "Transit delay selection and indication"},
  		{ 0x44, "Packet layer binary parameters"},
  		{ 0x45, "Packet layer window size"},
  		{ 0x46, "Packet size"},
  		{ 0x47, "Closed user group"},
  		{ 0x4a, "Reverse charge indication"},
  		{ 0x4c, "COLP"},
  		{ 0x6c, "Calling party number"}, 
  		{ 0x6d, "Calling party subaddress"},
  		{ 0x70, "Called party number"},
  		{ 0x71, "Called party subaddress"},
  		{ 0x74, "Redirecting number"},
  		{ 0x78, "Transit network selection"},
  		{ 0x79, "Restart indicator"},
  		{ 0x7c, "Low layer compatibility"},
  		{ 0x7d, "High layer compatibility"},
  		{ 0x7e, "User-user"},
  		{ 0x7f, "Escape for extension"},
};  

static const char* const ieunknown = "Unknown Info Element";

#define MTSIZE sizeof(mtlist)/sizeof(struct MessageType)
#define IESIZE sizeof(ielist)/sizeof(struct InformationElement)

const char* q931Pack::getMsgTyp(int x) {
	for (size_t i=0; i < MTSIZE; i++) {
		if (mtlist[i].nr == x)
			return mtlist[i].descr;
	}
	return mtlist[0].descr;
}

const char* q931Pack::getIETyp(int x) {
	for (size_t i=0; i < IESIZE; i++) {
		if (ielist[i].nr == x)
			return ielist[i].descr;
	}
	return ieunknown;
}

void q931Pack::prLocation(ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++;
	idx++;
	int loc = (ep->data[idx++] & 0x0f);
	
	switch (loc) {
		case 0x00 : strout << " User";                                break;
		case 0x01 : strout << " Private network serving local user";  break;
		case 0x02 : strout << " Public network serving local user";   break;
		case 0x03 : strout << " Transit network";                     break;
		case 0x04 : strout << " Public network serving remote user";  break;
		case 0x05 : strout << " Private network serving remote user"; break;
		case 0x07 : strout << " International network";               break;
		case 0x0a : strout << " Network beyond inter-working point";  break;
		default :	break; 
	}
}

void q931Pack::prDateTime(ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++; // Laenge
	int Y = ep->data[idx++];
	int M = ep->data[idx++];
	int D = ep->data[idx++];
	int h = ep->data[idx++];
	int m = ep->data[idx++];
	ostrstream s;
	s.form("%02d.%02d.%02d %02d:%02d",D,M,Y,h,m);
	s << ends;
	//DateTime = s.str();

	strout	<< "datetime: " << DateTime << endl;
}

void q931Pack::prUser2User(ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++;
	int len = ep->data[idx++] - 1;
	idx++;
	strout << " Text: ";
	while (len--) 
		strout << ep->data[idx++];
}

void q931Pack::prCalled (ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++;
	int len = ep->data[idx++] - 1;
	idx++;
	while (len--) 
		Called += ep->data[idx++];
	strout << " CalledPN: " << Called;
}

void q931Pack::prCalling (ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++;
	int len = ep->data[idx++] - 1;
	if (!(ep->data[idx] & 0x80)){
		idx++;
		len--;
	}
	idx++;
	while (len--) 
		Caller += ep->data[idx++];
	strout << " CallingPN: " << Caller;
	
}

void q931Pack::head(ESPMsgHdr *ep, ostrstream& strout) {
	if (debug > 1) strout << *ep << endl;
	if (ep->data[0] - 2) {				// First byte is direction
		strout << "network->user ";
		dir = DIR_NET2USER;
	} else {
		strout << "user->network ";
		dir = DIR_USER2NET;
	}
}

void q931Pack::analyzEuro(ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	int callreflen = ep->data[idx++];
	int callref = 0;
	if (callreflen)
		callref = ep->data[idx++];
	msgtype = ep->data[idx++];
	size_t ds = ep->data.size();
	

	strout << " CALLREFLEN: " << callreflen;
	strout << " CALLREF: " << (callref & 0x7F);
	if ((callref & 0x80))
		strout << " called ";
	else
		strout << " caller ";
	strout << " MSGTYPE: " << getMsgTyp(msgtype);
	strout << endl;
	while ((idx < ds)) {
		if ( ep->data[idx] & 0x80 ) {
			switch ((ep->data[idx] >> 4) & 7) {
				case 1:
					strout << "\tShift " << (int)(ep->data[idx] & 0x0F);
					break;
				case 3:
					strout << "\tCongestion level " << (int)(ep->data[idx] & 0x0F);
					break;
				case 5:
					strout << "\tRepeat indicator " << (int)(ep->data[idx] & 0x0F);
					break;
				case 2:
					if (ep->data[idx] == 0xA0) {
						strout << "\tMore data ";
						break;
					}
					if (ep->data[idx] == 0xA1) {
						strout << "\tSending complete ";
						break;
					}
				default:
					strout << "\tReserved " << (int)(ep->data[idx] & 0x0F);
					break;
			}
			idx++;
			continue;
		}

		strout << getIETyp(ep->data[idx]);
		strout << '[';
		strout << hex << (int)ep->data[idx];
		strout << "] Len: ";
		strout << hex << (int)(ep->data[idx + 1]);
		strout << dec;
		switch (ep->data[idx]) {
			case 0x1E:	// Progress indicator
				prLocation(ep, idx, strout);
				break;
			case 0x29:	// Progress indicator
				prDateTime(ep, idx, strout);
				break;
			case 0x4C:	// COLP
			case 0x6C:	// Calling Party Number
				prCalling(ep, idx, strout);
				break;
			case 0x70:	// Called Party Number
				prCalled(ep, idx, strout);
				break;
			case 0x7E:	// User2User-Msg
				prUser2User(ep, idx, strout);
				break;
		}
		strout << endl;
		idx += ep->data[idx + 1] + 2;
	}
}

void q931Pack::analyzAckermann(ESPMsgHdr *ep, size_t idx, ostrstream& strout) {
	idx++; //call ref len
	idx++; //call ref
	idx++; //msg type
	idx++; // Laenge
	msgtype = ep->data[idx++];
	switch(msgtype) {
		case 0x01: 
			{
				strout << "linestatus: " ;
				ack_empty = ep->data[idx++] - 5; // Laenge
				idx++; // 0x03
				idx++; // 0x03
				idx++; // 0x03
				int s0 = (int)ep->data[idx++];
				if ((s0 & 0x2) == 0x2) strout << "Int1a " ;
				if ((s0 & 0x1) == 0x1) strout << "Ext1a " ;
				s0 = (int)ep->data[idx++];
				if ((s0 & 0x2) == 0x2) strout << "Int1b " ;
				if ((s0 & 0x1) == 0x1) strout << "Ext1b " ;
				while (idx < ep->data.size()) {
					strout << hex << (int)ep->data[idx++] << ' ';
				}
				strout << dec ;
			}
			break;
		case 0x02:
			{
				idx++; // Laenge
				int Y = ep->data[idx++];
				int M = ep->data[idx++];
				int D = ep->data[idx++];
				int h = ep->data[idx++];
				int m = ep->data[idx++];
				ostrstream s;
				s.form("%02d.%02d.%02d %02d:%02d",D,M,Y,h,m);
				s << ends;
				DateTime = s.str();

				strout	<< "datetime: " << DateTime << endl;
			}
			break;
	}
}

void q931Pack::analyze(ESPMsgHdr *ep) {
	if (ep->data.size() < 3) return;	// Packets shouldn't be less than 3 bytes

	ostrstream strout;
	ack_empty = 0;

	size_t	ptr  = 2;
	BYTE	sapi = ep->data[ptr++] >> 2;
	BYTE	tei  = ep->data[ptr++] >> 1;
	BYTE	ftyp = ep->data[ptr++];

	if ((tei == GROUP_TEI) && (sapi == CTRL_SAPI) && (ftyp == 3)) {
		head(ep, strout);
		strout << "Broadcast ";
	} else if ((tei != GROUP_TEI) && (sapi == CTRL_SAPI) && (!(ftyp & 1))) {
		head(ep, strout);
		ptr++;
	} else 
		return;

	if (ep->data.size() <= ptr) {
		strout << "short frame" << endl;
		return;
	}

	strout << "sapi: " << (int)sapi << " tei: " << (int)tei << " ftyp: " << (int)ftyp;

	protocol = ep->data[ptr++];
	if (protocol == PROT_ACKERMANN) {
		strout << " proto: " << "Ackermann ";
		analyzAckermann(ep, ptr, strout);
	} else if (protocol == PROT_EURO) {
		strout << " proto: " << "EDSS1 ";
		analyzEuro(ep, ptr, strout);
	} else 
		strout << " proto: " << "IRGENDWAS";

	
	strout << endl;
	strout << ends;

	if ((msgtype == 1) && (protocol == PROT_ACKERMANN) && (!ack_empty)) {
		if (debug > 2) cout << strout.str();
	} else 
		if (debug) cout << strout.str();
}

